//--------------------------------------------------------------------
//      gen.cpp 
//      ------- 
//      Generator of the "sql" file with spatio-bitemporal data and queries
//      Implemented: TGenerator
//
//      Spatio-temporal data generator v 1.0 
//      Copyright(c) 1998-2000, Aalborg University
//

#include "gen.h"
#include "random.h"
#include "pool.h"
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#include <stdio.h>
#include <limits.h>
#include <macros.h>

bool gen_binary = true;

static int  logOps;      // number of logical index operations done so far
static int  numEntries;  // number of entries in the index
static int  numQueries;
static int  numFutureQueries;
static char op_code;     

//------------------------------------------------
TGenerator::TGenerator(TParams& params)
{
   total            = params.getInt  ("Total");
   initialInserts   = params.getInt  ("InitialInserts");
   probStart        = params.getFloat("ProbStart");
   probEnd          = params.getFloat("ProbEnd");
   probUpdInsert    = params.getFloat("ProbUpdInsert");
   probUpdDelete    = params.getFloat("ProbUpdDelete");
   devVt1           = params.getFloat("DevVt1");
   maxVtLength      = params.getInt  ("MaxVtLength");
   devDelta         = params.getFloat("DevDelta");
   statsStart       = params.getInt  ("StatsStart");
   statsInterval    = params.getInt  ("StatsInterval");
   updatesPerPoint  = params.getFloat("UpdatesPerPoint");
   beginning        = params.getInt  ("Beginning");
   pointSpData      = params.getInt  ("PointSpData"); 
   spChange         = params.getInt  ("SpChange");
   interval         = params.getInt  ("Interval");
   queryQty         = params.getInt  ("QueryQty");
   current          = params.getFloat("Current");
   char*  tn        = params.getString("Name");

   while (*tn == ' ') tn++;
   strcpy (table_name, tn);
   strcat (table_name, ".sql");

   if (params.Error())
   {
      cout << "Aborting." << endl;
      exit(1);
   }

   params.Silent();
   queryTempSpR     = params.getFloat("QueryTempSpR");
   queryVolume      = params.getFloat("QueryVolume");

   if (params.Error())
   {
      queryVolume = 0;
      params.ResetError();
      maxQWindowTT     = params.getInt  ("MaxQWindowTT");
      maxQWindowVT     = params.getInt  ("MaxQWindowVT");
      windows          = params.getFloat("Windows");
      slices           = params.getFloat("Slices");
      points           = params.getFloat("Points");
      maxQWindowSp     = params.getInt  ("MaxQWindowSp");
      probBitempQ      = params.getFloat("ProbBitempQ");            
      probSpatialQ     = params.getFloat("ProbSpatialQ");
      spPoints         = params.getFloat("SpPoints");
  
      if (params.Error())
      {
  	 cout << "GENERATOR Error: Not all query generation parameters are specified." 
              << endl << "Aborting." << endl;
         exit(1);
      }
   }
   else queryType = params.getInt ("QueryType");

   sqlf.open(table_name);

   if (!sqlf)
   {
      cout << "GENERATOR Error: Unable to open output file " << table_name << "."
           << endl;
      exit (1);
   }

   tn  = params.getString("SpatialData");
   if (!params.Error())
   {
      while (*tn == ' ') tn++;
      strcpy (spfile_name, tn);
      spdf.open(tn);
   }

   if (params.Error() || !spdf)
   {
      cout << "GENERATOR Error: Unable to open spatial data file " << tn 
           << "." << endl;
      exit(1);
   }
   else
      spdf >> minX >> minY >> maxX >> maxY;
}

//------------------------------------------------
static void OutTime (ostream& os, long time)
{
   if (gen_binary) 
   {
      op_code = OPC_TX;
      os.write(&op_code, 1);
      os.write(&time, sizeof(long));
   }
   else
      os << "t " << time << endl;
}

//------------------------------------------------
static void AdvTime (ostream& os, long time)
{
   if (gen_binary) 
   {
      op_code = time == 1 ? OPC_A : OPC_AX;
      os.write(&op_code, 1);  
      if (time != 1) os.write(&time, sizeof(long));
   }
   else
   {   
      os << "a";
      if (time != 1) os << ' ' << time;
      os << endl;
   }
}

//------------------------------------------------
static void OutStats (ostream& os, long time)
{
   if (gen_binary)
   {
      cout << "GENERATOR Warning: Out Stats are currently supported only in text mode. Skipping."
           << endl;
      return;  
   } 
   os << "output current time is " << time << endl;
   os << "stats out" << endl;
}

//------------------------------------------------
static void OutInsert (ostream& os, TSpBitemporal& key, unsigned short ptr)
{
   if (gen_binary)
   {
      key.write(os, OPC_I);
      os.write(&ptr, sizeof(short)); 
   }
   else 
      os << "i " << key << ',' << ptr << endl;
   logOps++;
   numEntries++;
}

//------------------------------------------------
static void OutSelect (ostream& os, TSpBitemporal& key)
{
   if (gen_binary)
      key.write(os, OPC_S);
   else 
      os << "s &" << key << endl;
}

//------------------------------------------------
static void OutDelete (ostream& os, TSpBitemporal& key)
{
   if (gen_binary)
      key.write(os, OPC_D);
   else 
      os << "d =" << key << endl;
   numEntries--;
}

//------------------------------------------------
void TGenerator::Generate ()
{
   TSpBitemporal  key;
   long           i, j, t, q = LONG_MAX;

   TPool  pool(initialInserts + (int)((total / 5 - initialInserts)*probStart));

   randomize();

   maxFutureVT = 0;
   maxKey      = 0;

   if (gen_binary) sqlf << "bin" << endl;

   assert(updatesPerPoint = 1.0);   // currently only this is supported
 
   OutTime (sqlf, beginning);

   cout << "Progress : 0%";
   numEntries = 0;
   numQueries = 0;
   numFutureQueries = 0; 
   for (i = 0, logOps = 0, t = beginning; logOps < total; i++, t++)
   {
      double coin = rnd();

      if (!(i % 100))
      {
         cout << "\rProgress : " << logOps * 100 / total << "%"; 
         cout.flush();
      } 

      if (i == initialInserts) q = logOps;
       
      if (logOps >= q)
      {
         GenerateQueries(t);
         q += interval;
      }

      if (i >= statsStart && !((i - statsStart)%statsInterval))
         OutStats(sqlf, t);
         
      if (i < initialInserts || coin < probStart)
         StartHistory(pool, t);       
      else if (coin < probStart + probEnd)
              EndHistory(pool, t);
           else UpdateHistory(pool, t);
         
      while (pool[pool.getNum() - 1]->futureEntry.tt1 == t) 
         ContinueHistory(pool, t);
        
      AdvTime(sqlf, 1);
   }   
   cout << "\rProgress : 100%" << endl;   

   cout << "Number of entries in the index (insertions - deletions): "
        << numEntries << endl;
   cout << "Percentage of future queries in the workload: "
        << (double) numFutureQueries / numQueries * 100 << '%' << endl;

   if (gen_binary) 
   { 
      op_code = OPC_EOF; 
      sqlf.write(&op_code, 1);
   }
}

//------------------------------------------------
static void LogicalDelete(ostream& os, TSpBitemporal& key, 
                          unsigned short ptr, long CT)
{
   OutDelete(os, key);
   key.tt2 = CT - 1;
   if (key.vt2 != NOW ||
       (key.vt1 > key.tt1 + key.delta ?
        key.vt1 - key.delta : key.tt1) <= key.tt2)
      OutInsert(os, key, ptr);
   else logOps++;
}

//------------------------------------------------
static void UpdateEnd(ostream& os, TSpBitemporal& key, long end, 
                      unsigned short ptr, long CT)
{
   LogicalDelete(os, key, ptr, CT);
   assert(end >= key.vt1);
   key.tt1   = CT;
   key.tt2   = UC;
   key.vt2   = end;
   key.delta = 0;
   OutInsert(os, key, ptr);
}

//------------------------------------------------
//  Starts valid time history of a changing spatial object
//
void TGenerator::StartHistory(TPool& pool, long CT)
{
   THistory*      hist  = new THistory;

   hist->futureEntry.tt1 = CT;
   hist->futureEntry.tt2 = UC;
   hist->futureEntry.vt1 = (long) randnorm (CT, devVt1);
   if (hist->futureEntry.vt1 < 0) hist->futureEntry.vt1 = 0;
   // vt2 and delta will be generated upon insertion of the futureEntry
   ReadSpatial(hist->futureEntry);
   hist->setLen(total / initialInserts / 3);
   hist->key = ++maxKey;
   hist->sp  = hist->futureEntry;

   pool.Add(hist);
   ContinueHistory(pool, CT);
}

//------------------------------------------------
//  Continues history by adding entries to the end and
//  scheduling the addition of a futureEntry
//
void TGenerator::ContinueHistory(TPool& pool, long CT)
{
   TSpBitemporal key;
   long          vtlen; 
   long          futureDevVt1;  
   THistory*     hist = pool[pool.getNum() - 1];

   assert(hist->futureEntry.tt1 == CT);

   // Ending the previous interval in the history (which was now-relative)
   if (!hist->isEmpty()) 
      UpdateEnd(sqlf, (*hist)[hist->getNum() - 1], 
                hist->futureEntry.vt1 - 1, hist->key, CT);   
   
   // Generating a difference between vt1 and tt1 of a new futureEntry
   futureDevVt1 = (long) randnorm (0, devVt1);  

   // Inserting current futureEntry and, if necessary, filling the gap
   key = hist->futureEntry;      
   for (;;)
   {
      // Generating a planned valid-time length 
      vtlen   = random(maxVtLength);    
      key.vt2 = key.vt1 + vtlen;
      if (key.vt2 >= CT + futureDevVt1)
      {
         key.delta = (long) randnorm (0, devDelta);
         key.vt2 = NOW;
      }
      else key.delta = 0;
      if (key.vt1 - CT > maxFutureVT) maxFutureVT = key.vt1 - CT;
      hist->Add(key);
      OutInsert(sqlf, key, hist->key);
      if (key.vt2 == NOW) break;
      key.vt1 = key.vt2 + 1;      // we stack them one on another         
      AdjustSpatial(key, key);
   }

   // Finishing generating a new futureEntry
   hist->futureEntry.vt1 = key.vt1 + vtlen + 1;
   hist->futureEntry.tt1 = hist->futureEntry.vt1 - futureDevVt1; 
   AdjustSpatial(hist->futureEntry, key);

   // Scheduling insertion of a new futureEntry
   pool.Schedule(pool.getNum() - 1);
}

//------------------------------------------------
void TGenerator::EndHistory (TPool& pool, long CT)
{         
   int       n    = pool.GetRandomAlive();
   if (n == pool.getNum()) return;   // Nothing to end, all are ended
   THistory* hist = pool[n];
 
   if (!hist->isEmpty())
   {
      TSpBitemporal& le = (*hist)[hist->getNum() - 1]; // last element

      assert(le.vt2 == NOW);
      if (le.vt1 <= CT + le.delta) 
         UpdateEnd(sqlf, le, CT + le.delta, hist->key, CT);   
      else 
      {
         LogicalDelete(sqlf, le, hist->key, CT);
         hist->Delete(hist->getNum() - 1);
      } 
   }
   
   // Schedule a history never to be continued
   hist->futureEntry.tt1 = LONG_MAX;
   pool.Schedule(n);
}

//------------------------------------------------
void TGenerator::UpdateHistory (TPool& pool, long CT)
{
   assert(pool.getNum());
   int           n    = random(pool.getNum());
   THistory*     hist = pool[n];

   if (hist->isEmpty()) return;  // Nothing to update, a chosen history is empty    
   if (rnd() < probUpdInsert)
   {  // Inserting a new valid-time interval in a past history 
      int           prev;
      long          endt;
      TSpBitemporal key;

      // Transaction time
      key.tt1 = CT;
      key.tt2 = UC;
      
      // Valid time
      endt = min((*hist)[hist->getNum() - 1].vtend(CT), hist->futureEntry.vt1 - 1);
      if (!endt) return;        // No place to insert anything 
      key.vt2 = (*hist)[0].vt1 - 1;
      if (key.vt2 < 0) key.vt2 = 0;
      key.vt2 += random (endt - key.vt2);
      key.vt1  = key.vt2 - random(maxVtLength);
      if (key.vt1 < 0) key.vt1 = 0;
      key.delta = 0;
      if (key.vt1 - CT > maxFutureVT) maxFutureVT = key.vt1 - CT;
      
      // Spatial part
      for (prev = hist->getNum() - 1; 
           prev && (*hist)[prev].vt1 >= key.vt1; prev--); 
      if ((*hist)[prev].vt1 < key.vt1)
         AdjustSpatial(key, (*hist)[prev]);
      else 
      { 
         AdjustSpatial(key, hist->sp);  // We generate before existing history 
         prev = -1;
      }

      // Correct all overlaping history intervals
      
      // if we cut the end of a previous interval
      if (prev != -1 && (*hist)[prev].vtend(CT) >= key.vt1) 
         if((*hist)[prev].vtend(CT) <= key.vt2)
            UpdateEnd(sqlf, (*hist)[prev], key.vt1 - 1, hist->key, CT);
         else 
	 {  // if we are fully overlaped by the "prev" interval  
            TSpBitemporal newpart = (*hist)[prev];

            UpdateEnd(sqlf, (*hist)[prev], key.vt1 - 1, hist->key, CT);
            newpart.vt1 = key.vt2 + 1;
            newpart.tt1 = CT;
            hist->Add(newpart, prev + 1); 
            OutInsert(sqlf, newpart, hist->key);   
         }
      prev++;

      // delete everything that we fully overlap
      while (prev < hist->getNum() && (*hist)[prev].vtend(CT) <= key.vt2)
      {
         LogicalDelete(sqlf, (*hist)[prev], hist->key, CT);
         hist->Delete(prev);
      }
     
      // if we cut the beginning of an interval
      if (prev < hist->getNum() && (*hist)[prev].vt1 <= key.vt2) 
      {
         LogicalDelete(sqlf, (*hist)[prev], hist->key, CT);
         (*hist)[prev].tt1 = CT;
         (*hist)[prev].tt2 = UC;
         (*hist)[prev].vt1 = key.vt2 + 1;
         OutInsert(sqlf, (*hist)[prev], hist->key);
      }   
      
      // Insert a new interval
      hist->Add(key, prev);
      OutInsert(sqlf, key, hist->key);
   }
   else  
   {  // Deleting an interval from a history 
      int j = random(hist->getNum());
      
      LogicalDelete(sqlf, (*hist)[j], hist->key, CT);
      if (j && (*hist)[j-1].vt2 == (*hist)[j].vt1 - 1)
      {  // Not leaving gaps in history 
         LogicalDelete(sqlf, (*hist)[j-1], hist->key, CT);
         (*hist)[j-1].tt1   = CT;
         (*hist)[j-1].tt2   = UC;
         (*hist)[j-1].vt2   = (*hist)[j].vt2;
         (*hist)[j-1].delta = (*hist)[j].delta;
         OutInsert(sqlf, (*hist)[j-1], hist->key);
      }
      hist->Delete(j);
   }
}

//------------------------------------------------
void TGenerator::ReadSpatial (TSpBitemporal& key)
{
   spdf >> key.x1 >> key.y1 >> key.x2;
   if (spdf.eof()) 
   {
      spdf.close();
      spdf.open (spfile_name);        
      // skip the minx, miny, maxx, maxy
      spdf >> key.x1 >> key.y1 >> key.x2 >> key.y2; 
      spdf >> key.x1 >> key.y1 >> key.x2 >> key.y2;
   }
   else spdf >> key.y2;
   if (pointSpData) 
   {
      key.x2 = key.x1;
      key.y2 = key.y1; 
   }
}

//------------------------------------------------
void TGenerator::AdjustSpatial (TSpBitemporal& to, TSpBitemporal& from)
{
   long deltaX = random(2*spChange + 1) - spChange; 
   long deltaY = random(2*spChange + 1) - spChange;
   if (pointSpData) 
   {
      to.x1 = from.x1 + deltaX;
      to.y1 = from.y1 + deltaY;
      to.x2 = to.x1;
      to.y2 = to.y1; 
   }
   else 
   {
      if (deltaX < 0 && deltaX < from.x1 - from.x2) deltaX = from.x1 - from.x2;
      if (deltaY < 0 && deltaY < from.y1 - from.y2) deltaY = from.y1 - from.y2;
      to.x1 = from.x1 - deltaX / 2;
      to.x2 = from.x2 + deltaX / 2;  
      to.y1 = from.y1 - deltaY / 2;
      to.y2 = from.y2 + deltaY / 2;  
   }
}

//------------------------------------------------
//  Two methods of query generation are implemented 
//  First of them preserves the exact given ratio between space and time
//
void TGenerator::GenerateQueries (long CT)
{
   TSpBitemporal  key;
   long           yttlen, xvtlen;
   double         coin, coin0, maxqv, volume, tarea, sarea;
   
   if (queryVolume != 0.0)
      for (long i = 0; i < queryQty; i++, numQueries++)
      { 
         if (queryType != 3)
	 {  
	    // maxqv ensures that an average ttlen is not larger 
            // than half of CT - beginning + 1
            double maxqv = pow(CT - beginning + 1, 4) * queryTempSpR / 8;
            volume       = randomf(queryVolume);
         }

         switch(queryType)
	 {
	  case 1:     // spatial point 
            maxqv = pow(CT - beginning + 1, 2) / 2;            
            tarea = randomf(maxqv < queryVolume ? maxqv : queryVolume);
            sarea = 1;
            break;
 
	  case 2:     // temporal point
            tarea = 1;
            sarea = randomf(queryVolume); 
            break;

	  case 3:     // transaction timeslice
            key.vt1 = 0;
            key.vt2 = CT + maxFutureVT + maxVtLength;
            yttlen = 0;
            sarea = 1;
            break;

  	  default:  // 0 - window
	    // maxqv ensures that an average ttlen is not larger 
            // than half of CT - beginning + 1
            maxqv  = pow(CT - beginning + 1, 4) * queryTempSpR / 8;
            volume = randomf(maxqv < queryVolume ? maxqv : queryVolume);
            tarea  = sqrt(volume / queryTempSpR);
            sarea  = volume / tarea;
         }
        
         // Temporal part
         if (queryType != 3)
	 {           
            yttlen = xvtlen = (long) sqrt (tarea) - 1;
            if (yttlen < 0) yttlen = 0;
            if (xvtlen < 0) xvtlen = 0;
         }

         if (rnd() < current) 
         {  
            key.tt2 = CT;
            key.tt1 = CT - yttlen;
            if (key.tt1 < beginning) key.tt1 = beginning;
         }   
         else 
  	 {
            if (yttlen > CT - beginning)
	    { 
               key.tt2 = CT;
               key.tt1 = beginning;
            } 
            else 
   	    { 
               key.tt2 = beginning + yttlen + random (CT + 1 - beginning - yttlen);
               key.tt1 = key.tt2 - yttlen;
            } 
         }
         if (queryType != 3)
	 {
	   //    if (rnd() < 0.5) key.vt1 = key.tt2 + 1 + random(maxFutureVT); 
           // else key.vt1 = random (key.tt2 + 1);
            key.vt1 = random (key.tt2 + 2*maxFutureVT);
            key.vt2 = key.vt1 + xvtlen;
            if (key.vt1 > key.tt2) numFutureQueries++;
         }
         key.delta = NO_DELTA;
         assert(key.tt1 <= key.tt2);
         assert(key.vt1 <= key.vt2);
         
         // Spatial part
         yttlen = xvtlen = (long) sqrt (sarea) - 1; 
         if (yttlen < 0) yttlen = 0;
         if (xvtlen < 0) xvtlen = 0;
         
         if (yttlen > maxY - minY)
	 { 
            key.y2 = maxY;
            key.y1 = minY;
         } 
         else 
   	 { 
            key.y2 = minY + yttlen + random (maxY + 1 - minY - yttlen);
            key.y1 = key.y2 - yttlen;
         }          
         if (xvtlen > maxX - minX)
	 { 
            key.x2 = maxX;
            key.x1 = minX;
         } 
         else 
         { 
            key.x2 = minX + xvtlen + random (maxX + 1 - minX - xvtlen);
            key.x1 = key.x2 - xvtlen;
         }          
            
         OutSelect(sqlf, key);
      } 
   else  
      for (long i = 0; i < queryQty; i++)
      { 
         coin0 = rnd();

         if (coin0 < probBitempQ || coin0 >= probBitempQ + probSpatialQ)
         {
            coin = rnd();
            yttlen = coin < windows ? random (maxQWindowTT) : 0;
   
            if (rnd() < current) 
            {  
               key.tt2 = CT;
               key.tt1 = CT - yttlen;
               if (key.tt1 < beginning) key.tt1 = beginning;
            }   
            else 
  	    {
               if (yttlen > CT - beginning)
	       { 
                  key.tt2 = CT;
                  key.tt1 = beginning;
               } 
               else 
   	       { 
                   key.tt2 = beginning + yttlen + random (CT + 1 - beginning - yttlen);
                   key.tt1 = key.tt2 - yttlen;
               } 
            }
/*          if (key.tt2 > qwBottom)
               key.vt1 = key.vt2 = key.tt2 - qwBottom + 
                                   random (qwBottom + maxFutureVT / 2); 
            else
	       key.vt1 = key.vt2 = random (key.tt2 + maxFutureVT / 2);
*/
            key.vt1 = key.vt2 = random (key.tt2 + maxFutureVT);
            if (coin < windows + slices)
               key.vt2 += (long) random (maxQWindowVT);
            key.delta = NO_DELTA;
            assert(key.tt1 <= key.tt2);
            assert(key.vt1 <= key.vt2);
         }
         else 
         {  // Whole bitemporal space  
            key.tt1 = 0;  
            key.tt2 = CT;
            key.vt1 = 0;
            key.vt2 = CT + 1000000; 
            key.delta = NO_DELTA;
         }

    
         if (coin0 >= probBitempQ)
         { 
            key.x1 = key.x2 = minX + random (maxX - minX + 1);
            key.y1 = key.y2 = minY + random (maxY - minY + 1);
            if (rnd() > spPoints)
            {
               key.x2 +=  random (maxQWindowSp);
               key.y2 +=  random (maxQWindowSp);
            }
         }
         else 
         {
            key.x1 = minX;
            key.x2 = maxX;
            key.y1 = minY;
            key.y2 = maxY;
         } 

         OutSelect(sqlf, key);
      }
}

