//--------------------------------------------------------------------
//      pool.h 
//      ------ 
//      Pool of the spatio-bitemporal data objects
//      Defined: TSpBitemporal, THistory, TPool
//
//      Spatio-temporal data generator v 1.0 
//      Copyright(c) 1998-2000, Aalborg University
//

#ifndef __POOL_H
#define __POOL_H

#include <iostream.h>
#include <stddef.h>
#include "random.h"
#include "grx_opcodes.h"

#define UC         (-1)
#define NOW        (-2)
#define NO_DELTA   0x7FFFFFFF    // For keys in search predicates
#define NO_SPATIAL 0x7FFFFFFF    // For keys without spatial extent

#define POOL_INC     1000
#define HISTORY_INC  5

struct TSpBitemporal
{
   long  tt1, tt2;
   long  vt1, vt2;
   long  x1,  x2;      // x1 can hold special value NO_SPATIAL
   long  y1,  y2;
   long  delta;        // can hold special value NO_DELTA

   long  vtend(long CT);
   
   ostream& write(ostream& os, char opc); 
 
   friend int operator== (TSpBitemporal& bt1, TSpBitemporal& bt2);
   friend ostream& operator << (ostream& os, TSpBitemporal& bt);
};

inline long TSpBitemporal::vtend(long CT) 
{ return vt2 == NOW ? (CT + delta > vt1 ? CT + delta : vt1) : vt2; }

inline int operator== (TSpBitemporal& bt1, TSpBitemporal& bt2)
{
   return bt1.tt1 == bt2.tt1 && bt1.tt2 == bt2.tt2 &&
          bt1.vt1 == bt2.vt1 && bt1.vt2 == bt2.vt2 &&
          bt1.x1  == bt2.x1  && bt1.x2  == bt2.x2  &&
          bt1.y1  == bt2.y1  && bt1.y2  == bt2.y2  &&
          bt1.delta == bt2.delta;
}

inline ostream& TSpBitemporal::write(ostream& os, char opc)
{
   if (tt2 != UC && vt2 != NOW)
   {
      os.write(&opc, 1);
      os.write(&tt1, sizeof(long));
      os.write(&vt1, sizeof(long));
      os.write(&tt2, sizeof(long));
      os.write(&vt2, sizeof(long));
   }
   else if (tt2 == UC && vt2 != NOW)
   {
      opc++;
      os.write(&opc, 1);
      os.write(&tt1, sizeof(long));
      os.write(&vt1, sizeof(long));
      os.write(&vt2, sizeof(long));
   }
   else if (tt2 != UC && vt2 == NOW)
   {
      opc += 2;
      os.write(&opc, 1);
      os.write(&tt1, sizeof(long));
      os.write(&vt1, sizeof(long));
      os.write(&tt2, sizeof(long));
      if (delta != NO_DELTA) os.write(&delta, sizeof(long));
   }
   else
   {
      opc += 3;
      os.write(&opc, 1);
      os.write(&tt1, sizeof(long));
      os.write(&vt1, sizeof(long));
      if (delta != NO_DELTA) os.write(&delta, sizeof(long));
   }

   if (x1 != NO_SPATIAL)
   {
      os.write(&x1, sizeof(long));
      os.write(&y1, sizeof(long));
      os.write(&x2, sizeof(long));
      os.write(&y2, sizeof(long));
   }
}

inline ostream& operator << (ostream& os, TSpBitemporal& bt)
{
   os << '(' << bt.tt1 << ',' << bt.vt1 << ',';
   if (bt.tt2 == UC) os << "UC";
   else              os << bt.tt2;
   os << ',';
   if (bt.vt2 == NOW) os << "NOW";
   else               os << bt.vt2;
   if (bt.delta != NO_DELTA) os << ',' << bt.delta;
   if (bt.x1 != NO_SPATIAL)
      os << ',' << bt.x1 << ',' << bt.y1 << ',' << bt.x2 << ',' << bt.y2;
   os << ')';

   return os;
}

//------------------------------------------------
// THistory - is a container of all TSpBitemporal 
//            objects, representing a valid time history 
//            of an entity
//
class THistory
{
 public:
   THistory()      : array(NULL), len(0), num(0), key(0) {}
   THistory(int n) : array(NULL), len(0), num(0), key(0) { setLen(n); }
   ~THistory();

   void setLen (int n);
   int  getNum () const {return  num; }
   int  isEmpty() const {return !num; }

   TSpBitemporal& operator[] (int n) const;

   void  Add    (TSpBitemporal& elem, int n = -1); 
   void  Delete (int n);

   unsigned short  key;          // identifier of the entity 
   TSpBitemporal   sp;           // to keep "original" spatial part
   TSpBitemporal   futureEntry;  // the entry that will be inserted at futureEntry.tt1 
     
 protected:
   int            len;
   int            num;
   TSpBitemporal* array;    
};

//------------------------------------------------
// TPool - is an array of THistory's
// TPool "owns" THistories, i.e., it deletes them in Delete and ~TPool
//
class TPool
{
 public:
   TPool()      : array(NULL), len(0), num(0), numEnded(0) {}
   TPool(int n) : array(NULL), numEnded(0)                 { setLen(n); }
   ~TPool();

   void setLen (int n);
   int  getNum () const { return num; }

   THistory* operator[] (int n) const;

   void Add     (THistory* hist);
   void Schedule(int n);
   void Delete  (int n);
   int  GetRandomAlive() { return numEnded + random (num - numEnded); }

 protected:
   int           len;
   int           num;
   int           numEnded;
   THistory**    array;
};

#endif


